searchNodes=[{"doc":"Sage is a dependency-free implementation of Sagas pattern in pure Elixir. It is a go to way when you dealing with distributed transactions, especially with an error recovery/cleanup. Sage does it's best to guarantee that either all of the transactions in a saga are successfully completed or compensating that all of the transactions did run to amend a partial execution.This is done by defining two way flow with transaction and compensation functions. When one of the transactions fails, Sage will ensure that transaction's and all of it's predecessors compensations are executed. However, it's important to note that Sage can not protect you from a node failure that executes given Sage.Critical Error HandlingFor TransactionsTransactions are wrapped in a try..catch block.Whenever a critical error occurs (exception is raised, error thrown or exit signal is received) Sage will run all compensations and then reraise the exception with the same stacktrace, so your log would look like it occurred without using a Sage.For CompensationsBy default, compensations are not protected from critical errors and would raise an exception. This is done to keep simplicity and follow &quot;let it fall&quot; pattern of the language, thinking that this kind of errors should be logged and then manually investigated by a developer.But if that's not enough for you, it is possible to register handler via with_compensation_error_handler/2. When it's registered, compensations are wrapped in a try..catch block and then it's error handler responsibility to take care about further actions. Few solutions you might want to try:Send notification to a Slack channel about need of manual resolution;Retry compensation;Spawn a new supervised process that would retry compensation and return an error in the Sage. (Useful when you have connection issues that would be resolved at some point in future.)Logging for compensation errors is pretty verbose to drive the attention to the problem from system maintainers.finally/2 hookSage does it's best to make sure final callback is executed even if there is a program bug in the code. This guarantee simplifies integration with a job processing queues, you can read more about it at GenTask Readme.If an error is raised within finally/2 hook, it's getting logged and ignored. Follow the simple rule - everything that is on your critical path should be a Sage transaction.Tracing and measuring Sage execution stepsSage allows you to set a tracer module which is called on each step of the execution flow (before and after transactions and/or compensations). It could be used to report metrics on the execution flow.If error is raised within tracing function, it's getting logged and ignored.","ref":"Sage.html","title":"Sage","type":"module"},{"doc":"Executes a Sage.Optionally, you can pass global options in opts, that will be sent to all transaction, compensation functions and hooks. It is especially useful when you want to have keep sage definitions declarative and execute them with different arguments (eg. you may build your Sage struct in a module attribute, because there is no need to repeat this work for each execution).If there was an exception, throw or exit in one of transaction functions, Sage will reraise it after compensating all effects.For handling exceptions in compensation functions see &quot;Critical Error Handling&quot; in module doc.Raises Sage.EmptyError if Sage does not have any transactions.","ref":"Sage.html#execute/2","title":"Sage.execute/2","type":"function"},{"doc":"Appends the Sage with a function that will be triggered after Sage execution.Registering duplicated final hook is not allowed and would raise an Sage.DuplicateFinalHookError exception.For hook specification see final_hook/0.","ref":"Sage.html#finally/2","title":"Sage.finally/2","type":"function"},{"doc":"Creates a new Sage.","ref":"Sage.html#new/0","title":"Sage.new/0","type":"function"},{"doc":"Appends sage with a transaction that does not have side effect.This is an alias for calling run/4 with a :noop instead of compensation callback.","ref":"Sage.html#run/3","title":"Sage.run/3","type":"function"},{"doc":"Appends sage with a transaction and function to compensate it's effect.Raises Sage.DuplicateStageError exception if stage name is duplicated for a given sage.CallbacksCallbacks can be either anonymous function or an {module, function, [arguments]} tuple. For callbacks interface see transaction/0 and compensation/0 type docs.Noop compensationIf transaction does not produce effect to compensate, pass :noop instead of compensation callback or use run/3.","ref":"Sage.html#run/4","title":"Sage.run/4","type":"function"},{"doc":"Appends sage with an asynchronous transaction and function to compensate it's effect.Asynchronous transactions are awaited before the next synchronous transaction or in the end of sage execution. If there is an error in asynchronous transaction, Sage will await for other transactions to complete or fail and then compensate for all the effect created by them.CallbacksTransaction callback for asynchronous stages receives only effects created by preceding synchronous transactions.For more details see run/4.Options:timeout - the time in milliseconds to wait for the transaction to finish, :infinity will wait indefinitely (default: 5000);","ref":"Sage.html#run_async/5","title":"Sage.run_async/5","type":"function"},{"doc":"Executes Sage with Ecto.Repo.transaction/2.Transaction is rolled back on error.Ecto must be included as application dependency if you want to use this function.","ref":"Sage.html#transaction/4","title":"Sage.transaction/4","type":"function"},{"doc":"Register error handler for compensations.Adapter must implement Sage.CompensationErrorHandler behaviour.For more information see &quot;Critical Error Handling&quot; in the module doc.","ref":"Sage.html#with_compensation_error_handler/2","title":"Sage.with_compensation_error_handler/2","type":"function"},{"doc":"Registers tracer for a Sage execution.Registering duplicated tracing callback is not allowed and would raise an Sage.DuplicateTracerError exception.All errors during execution of a tracing callbacks would be logged, but it won't affect Sage execution.Tracing module must implement Sage.Tracer behaviour. For more information see Sage.Tracer.handle_event/3.","ref":"Sage.html#with_tracer/2","title":"Sage.with_tracer/2","type":"function"},{"doc":"Options for asynchronous transaction stages.","ref":"Sage.html#t:async_opts/0","title":"Sage.async_opts/0","type":"type"},{"doc":"Compensation callback, can either anonymous function or an {module, function, [arguments]} tuple.Receives:effect created by transaction it's responsible for or nil in case effect is not known due to an error;effects created by preceding executed transactions;options passed to execute/2 function.Returns::ok if effect is compensated, Sage will continue to compensate other effects;:abort if effect is compensated but should not be created again, Sage will compensate other effects and ignore retries on any stages;{:retry, retry_opts} if effect is compensated but transaction can be retried with options retry_opts;{:continue, effect} if effect is compensated and execution can be retried with other effect to replace the transaction return. This allows to implement circuit breaker.Circuit BreakerAfter receiving a circuit breaker response Sage will continue executing transactions by using returned effect.Circuit breaking is only allowed if compensation function that returns it is responsible for the failed transaction (they both are parts of for the same execution step). Otherwise curcuit breaker would be ignored and Sage will continue applying backward recovery.The circuit breaker should use data which is local to the sage execution, preferably from list of options which are set via execute/2 2nd argument. This would guarantee that circuit breaker would not fail when response cache is not available.RetriesAfter receiving a {:retry, [retry_limit: limit]} Sage will retry the transaction on a stage where retry was received.Take into account that by doing retires you can increase execution time and block process that executes the Sage, which can produce timeout, eg. when you trying to respond to an HTTP request.Compensation guidelinesGeneral rule is that irrespectively to what compensate wants to return, effect must be always compensated. No matter what, side effects must not be created from compensating transaction.A compensating transaction doesn't necessarily return the data in the system to the state it was in at the start of the original operation. Instead, it compensates for the work performed by the steps that completed successfully before the operation failed.source: https://docs.microsoft.com/en-us/azure/architecture/patterns/compensating-transactionYou should try to make your compensations idempotent, which makes possible to retry if compensating transaction itself fails. According a modern HTTP semantics, the PUT and DELETE verbs are idempotent. Also, some services support idempotent requests via idempotency keys.Compensation transactions should not rely on effects created by preceding executed transactions, otherwise it will be more likely that your code is not idempotent and harder to maintain. Use them only as a last resort.","ref":"Sage.html#t:compensation/0","title":"Sage.compensation/0","type":"type"},{"doc":"Effects created on Sage execution.","ref":"Sage.html#t:effects/0","title":"Sage.effects/0","type":"type"},{"doc":"Final hook.It receives :ok if all transactions are successfully completed or :error otherwise and options passed to the execute/2.Return is ignored.","ref":"Sage.html#t:final_hook/0","title":"Sage.final_hook/0","type":"type"},{"doc":"Retry options.Retry count for all a sage execution is shared and stored internally, so even trough you can increase retry limit - retry count would be never reset to make sure that execution would not be retried infinitely.Available retry options::retry_limit - is the maximum number of possible retry attempts;:base_backoff - is the base backoff for retries in ms, no backoff is applied if this value is nil or not set;:max_backoff - is the maximum backoff value, default: 5_000 ms.;:enable_jitter - whatever jitter is applied to backoff value, default: true;Sage will log and give up retrying if options are invalid.Backoff calculationFor exponential backoff this formula is used:min(max_backoff, (base_backoff * 2) ^ retry_count)Example:AttemptBase BackoffMax BackoffSleep time11030000202103000040031030000800041030000300005103000030000When jitter is enabled backoff value is randomized:random(0, min(max_backoff, (base_backoff * 2) ^ retry_count))Example:AttemptBase BackoffMax BackoffSleep interval110300000..20210300000..400310300000..8000410300000..30000510300000..30000For more reasoning behind using jitter, check out this blog post.","ref":"Sage.html#t:retry_opts/0","title":"Sage.retry_opts/0","type":"type"},{"doc":"Name of Sage execution stage.","ref":"Sage.html#t:stage_name/0","title":"Sage.stage_name/0","type":"type"},{"doc":"","ref":"Sage.html#t:t/0","title":"Sage.t/0","type":"type"},{"doc":"Transaction callback, can either anonymous function or an {module, function, [arguments]} tuple.Receives effects created by preceding executed transactions and options passed to execute/2 function.Returns {:ok, effect} if transaction is successfully completed, {:error, reason} if there was an error or {:abort, reason} if there was an unrecoverable error. On receiving {:abort, reason} Sage will compensate all side effects created so far and ignore all retries.Sage.MalformedTransactionReturnError is raised after compensating all effects if callback returned malformed result.Transaction guidelinesYou should try to make your transactions idempotent, which makes possible to retry if compensating transaction itself fails. According a modern HTTP semantics, the PUT and DELETE verbs are idempotent. Also, some services support idempotent requests via idempotency keys.","ref":"Sage.html#t:transaction/0","title":"Sage.transaction/0","type":"type"},{"doc":"This module provides behaviour for compensation error handling.Few solutions you might want to try:Send notification to a Slack channel about need of manual resolution;Retry compensation;Spin off a new supervised process that would retry compensation and return an error in the Sage. (Useful when you have connection issues that would be resolved at some point in future.)For more information see &quot;Critical Error Handling&quot; in Sage module doc.","ref":"Sage.CompensationErrorHandler.html","title":"Sage.CompensationErrorHandler","type":"behaviour"},{"doc":"Handler for critical errors for compensation execution.It should return only {:error, reason} which would be returned by the Sage itself.","ref":"Sage.CompensationErrorHandler.html#c:handle_error/3","title":"Sage.CompensationErrorHandler.handle_error/3","type":"callback"},{"doc":"","ref":"Sage.CompensationErrorHandler.html#t:compensations_to_run/0","title":"Sage.CompensationErrorHandler.compensations_to_run/0","type":"type"},{"doc":"","ref":"Sage.CompensationErrorHandler.html#t:error/0","title":"Sage.CompensationErrorHandler.error/0","type":"type"},{"doc":"This module is responsible for Sage execution.","ref":"Sage.Executor.html","title":"Sage.Executor","type":"module"},{"doc":"","ref":"Sage.Executor.html#execute/2","title":"Sage.Executor.execute/2","type":"function"},{"doc":"","ref":"Sage.Executor.html#t:state/0","title":"Sage.Executor.state/0","type":"type"},{"doc":"This module implements retry logic with exponential back-off for compensations that want to retry transaction.","ref":"Sage.Executor.Retries.html","title":"Sage.Executor.Retries","type":"module"},{"doc":"Returns true if transaction should be retried, false - otherwise. Optionally, return would be delayed with an exponential backoff based on Sage.retry_opts/0.Malformed retry options would be logged and ignored.","ref":"Sage.Executor.Retries.html#retry_with_backoff?/2","title":"Sage.Executor.Retries.retry_with_backoff?/2","type":"function"},{"doc":"This module provides behaviour for Sage tracers.Tracing module is called before and after each transaction or compensation. For asynchronous operations this hook is triggered after awaiting for started transactions.Hooks StateAll hooks share their state, which by default contains options passed to execute/2 function. This is useful if you want to persist timer of stage execution start and then persist it somewhere.Altering this state won't affect transactions or compensations in any way, changes would be visible only to other tracing calls.","ref":"Sage.Tracer.html","title":"Sage.Tracer","type":"behaviour"},{"doc":"Handler for Sage execution event.It receives name of Sage execution stage, type of event (see action/0) and state shared for all tracing calls (see &quot;Hooks State&quot; in module doc).Returns updated state.","ref":"Sage.Tracer.html#c:handle_event/3","title":"Sage.Tracer.handle_event/3","type":"callback"},{"doc":"","ref":"Sage.Tracer.html#t:action/0","title":"Sage.Tracer.action/0","type":"type"},{"doc":"Raised at runtime when the asynchronous transaction timed out.","ref":"Sage.AsyncTransactionTimeoutError.html","title":"Sage.AsyncTransactionTimeoutError","type":"exception"},{"doc":"Raised at runtime when duplicated final hook is added to Sage.","ref":"Sage.DuplicateFinalHookError.html","title":"Sage.DuplicateFinalHookError","type":"exception"},{"doc":"Raised at runtime when operation with duplicated name is added to Sage.","ref":"Sage.DuplicateStageError.html","title":"Sage.DuplicateStageError","type":"exception"},{"doc":"Raised at runtime when a duplicated tracer is added to Sage.","ref":"Sage.DuplicateTracerError.html","title":"Sage.DuplicateTracerError","type":"exception"},{"doc":"Raised at runtime when empty sage is executed.","ref":"Sage.EmptyError.html","title":"Sage.EmptyError","type":"exception"},{"doc":"Raised at runtime when the compensation or operation has an malformed return.","ref":"Sage.MalformedCompensationReturnError.html","title":"Sage.MalformedCompensationReturnError","type":"exception"},{"doc":"Raised at runtime when the transaction or operation has an malformed return.","ref":"Sage.MalformedTransactionReturnError.html","title":"Sage.MalformedTransactionReturnError","type":"exception"},{"doc":"SageSage is a dependency-free implementation of the Sagas pattern in pure Elixir and provides a set of additional built-in features.It is a go-to way for dealing with distributed transactions, especially with an error recovery/cleanup. Sage does it's best to guarantee that either all of the transactions in a saga are successfully completed, or compensating that all of the transactions did run to amend a partial execution.It’s like Ecto.Multi but across business logic and third-party APIs.-- &lt;cite&gt;@jayjun&lt;/cite&gt;This is done by defining two way flow with transaction and compensation functions. When one of the transactions fails, Sage will ensure that the transaction's and all of its predecessors' compensations are executed. However, it's important to note that Sage can not protect you from a node failure that executes given Sage.To visualize it, let's imagine we have a 4-step transaction. Successful execution flow would look like:[T1] -&gt; [T2] -&gt; [T3] -&gt; [T4]and if we get a failure on 3-d step, Sage would cleanup side effects by running compensation functions:[T1] -&gt; [T2] -&gt; [T3 has an error] ↓ [C1] &lt;- [C2] &lt;- [C3]","ref":"readme.html","title":"Sage","type":"extras"},{"doc":"Along with that simple idea, you will get much more out of the box with Sage:Transaction retries;Asynchronous transactions with timeout;Retries with exponential backoff and jitter;Ease to write circuit breakers;Code that is clean and easy to test;Low cost of integration in existing code base and low performance overhead;Ability to not lock the database with long running transactions;Extensibility - write your own handler for critical errors or metric collector to measure how much time each step took.","ref":"readme.html#additional-features","title":"Sage - Additional Features","type":"extras"},{"doc":"Become a de facto tool to run distributed transactions in the Elixir world;Stay simple to use and small to maintain: less code - less bugs;Educate people how to run distributed transaction pragmatically.","ref":"readme.html#goals","title":"Sage - Goals","type":"extras"},{"doc":"Lot's of applications I've seen face a common task - interaction with third-party API's to offload some of the work on SaaS products or micro-services, when you simply need to commit to more than one database or in all other cases where you don't have transaction isolation between business logic steps (that we all got used to thanks to RDBMS).When dealing with those, it is a common desire to handle all sorts of errors when application code has failed in the middle of a transaction so that you won't leave databases in an inconsistent state.Using with (the old way)One solution is to write business logic using with syntax. But when the number of transaction steps grow, code becomes hard to maintain, test and even looks ugly. Consider the following pseudo-code (don't do this):defmodule WithExample do def create_and_subscribe_user(attrs) do Repo.transaction(fn -&gt; with {:ok, user} &lt;- create_user(attrs), {:ok, plans} &lt;- fetch_subscription_plans(attrs), {:ok, charge} &lt;- charge_card(user, subscription), {:ok, subscription} &lt;- create_subscription(user, plan, attrs), {:ok, _delivery} &lt;- schedule_delivery(user, subscription, attrs), {:ok, _receipt} &lt;- send_email_receipt(user, subscription, attrs), {:ok, user} &lt;- update_user(user, %{subscription: subscription}) do acknowledge_job(opts) else {:error, {:charge_failed, _reason}} -&gt; # First problem: charge is not available here :ok = refund(charge) reject_job(opts) {:error, {:create_subscription, _reason}} -&gt; # Second problem: growing list of compensations :ok = refund(charge) :ok = delete_subscription(subscription) reject_job(opts) # Third problem: how to decide when we should be sending another email or # at which stage we&#39;ve failed? other -&gt; # Will rollback transaction on all other errors :ok = ensure_deleted(fn -&gt; refund(charge) end) :ok = ensure_deleted(fn -&gt; delete_subscription(subscription) end) :ok = ensure_deleted(fn -&gt; delete_delivery_from_schedule(delivery) end) reject_job(opts) other end end) end defp ensure_deleted(cb) do case cb.() do :ok -&gt; :ok {:error, :not_found} -&gt; :ok end end endAlong with the issues highlighted in the code itself, there are few more:To know at which stage we failed we need to keep an eye on the special returns from the functions we're using here;Hard to control that there is a condition to compensate for all possible error cases;Impossible to not keep relative code close to each other, because bare expressions in withdo not leak to the else block;Hard to test;Hard to improve, eg. it is hard to add retries, async operations or circuit breaker without making it even worse.For some time you might get away by splitting create_and_subscribe_user/1, but it only works while the number of transactions is very small.Using SagasInstead, let's see how that pipeline would look with Sage:defmodule SageExample do import Sage require Logger @spec create_and_subscribe_user(attrs :: map()) :: {:ok, last_effect :: any(), all_effects :: map()} | {:error, reason :: any()} def create_and_subscribe_user(attrs) do new() |&gt; run(:user, &amp;create_user/2) |&gt; run(:plans, &amp;fetch_subscription_plans/2, &amp;subscription_plans_circuit_breaker/3) |&gt; run(:subscription, &amp;create_subscription/2, &amp;delete_subscription/3) |&gt; run_async(:delivery, &amp;schedule_delivery/2, &amp;delete_delivery_from_schedule/3) |&gt; run_async(:receipt, &amp;send_email_receipt/2, &amp;send_excuse_for_email_receipt/3) |&gt; run(:update_user, &amp;set_plan_for_a_user/2) |&gt; finally(&amp;acknowledge_job/2) |&gt; transaction(SageExample.Repo, attrs) end # Transaction behaviour: # @callback transaction(attrs :: map()) :: {:ok, last_effect :: any(), all_effects :: map()} | {:error, reason :: any()} # Compensation behaviour: # @callback compensation( # effect_to_compensate :: any(), # effects_so_far :: map(), # attrs :: any() # ) :: :ok | :abort | {:retry, retry_opts :: Sage.retry_opts()} | {:continue, any()} def create_user(_effects_so_far, %{&quot;user&quot; =&gt; user_attrs}) do %SageExample.User{} |&gt; SageExample.User.changeset(user_attrs) |&gt; SageExample.Repo.insert() end def fetch_subscription_plans(_effects_so_far, _attrs) do {:ok, _plans} = SageExample.Billing.APIClient.list_plans() end # If we failed to fetch plans, let&#39;s continue with cached ones def subscription_plans_circuit_breaker(_effect_to_compensate, _effects_so_far, _attrs) do {:continue, [%{&quot;id&quot; =&gt; &quot;free&quot;, &quot;total&quot; =&gt; 0}, %{&quot;id&quot; =&gt; &quot;standard&quot;, &quot;total&quot; =&gt; 4.99}]} end def create_subscription(%{user: user}, %{&quot;subscription&quot; =&gt; subscription}) do {:ok, subscription} = SageExample.Billing.APIClient.subscribe_user(user, subscription[&quot;plan&quot;]) end def delete_subscription(_effect_to_compensate, %{user: user}, _attrs) do :ok = SageExample.Billing.APIClient.delete_all_subscriptions_for_user(user) # We want to apply forward compensation from :subscription stage for 5 times {:retry, retry_limit: 5, base_backoff: 10, max_backoff: 30_000, enable_jitter: true} end # .. other transaction and compensation callbacks def acknowledge_job(:ok, attrs) do Logger.info(&quot;Successfully created user \#{attrs[&quot;user&quot;][&quot;email&quot;]}&quot;) end def acknowledge_job(_error, attrs) do Logger.warn(&quot;Failed to create user \#{attrs[&quot;user&quot;][&quot;email&quot;]}&quot;) end endAlong with a readable code, you are getting:Reasonable guarantees that all transaction steps are completed or all failed steps are compensated;Code which is much simpler and easier to test a code;Retries, circuit breaking and asynchronous requests our of the box;Declarative way to define your transactions and run them.Testing is easier, because instead of one monstrous function you will have many small callbacks which are easy to cover with unit tests. You only need to test business logic in transactions and that compensations are able to cleanup their effects. Sage itself has 100% test coverage.Even more, it is possible to apply a new kind of architecture in an Elixir project where Phoenix contexts (or just application domains) are providing helper functions for building sagas to a controller, which then uses one or more of them to make sure that each request is side-effects free. Simplified example:defmodule SageExample.UserController do use SageExample.Web, :controller action_fallback SageExample.FallbackController def signup_and_accept_team_invitation(conn, attrs) do Sage.new() |&gt; SageExample.Users.Sagas.create_user() |&gt; SageExample.Teams.Sagas.accept_invitation() |&gt; SageExample.Billing.Sagas.prorate_team_size() |&gt; Sage.execute(attrs) end endIf you want to have more examples, check out this blog post on Sage.","ref":"readme.html#rationale-use-cases","title":"Sage - Rationale (use cases)","type":"extras"},{"doc":"While Sage will do its best to compensate failures in a transaction and leave a system in a consistent state, there are some edge cases where it wouldn't be possible.What if my transaction has bugs or other errors? Transactions are wrapped in a try..catch block and would tolerate any exception, exit or rescue. And after executing compensations, an error will be reraised.What if my compensation has bugs or other errors? By default, compensations would not try to handle any kinds of errors. But you can write an adapter to handle those. For more information see Critical Error Handling section.What if the process that executes Sage or whole node fails? Right now Sage doesn't provide a way to tolerate failures of executing processes. (However, there is an RFC that aims for that.)What if an external API call fails and it's impossible to revert a step? In such cases, the process which is handling the pipeline will crash and the exception will be thrown. Make sure that you have a way of reacting to such cases (in some cases it might be acceptable to ignore the error while others might require a manual intervention).Can I be absolutely sure that everything went well? Unfortunately, no. As with any other distributed system, messages could be lost, the network could go down, hardware could fail etc. There is no way to programmatically solve all those cases, even retrying compensations won't help in some of such cases.For example, it's possible that a reply from an external API is lost even though a request actually succeeded. In such cases, you might want to retry the compensation which might have an unexpected result. Best way to solve that issue is to write compensations in an idempotent way and to always make sure that you have proper monitoring tools in place.","ref":"readme.html#execution-guarantees-and-edge-cases","title":"Sage - Execution Guarantees and Edge Cases","type":"extras"},{"doc":"For TransactionsTransactions are wrapped in a try..catch block.Whenever a critical error occurs (exception is raised, error thrown or exit signal is received) Sage will run all compensations and then reraise the exception with the same stacktrace, so your log would look like it occurred without using a Sage.For CompensationsBy default, compensations are not protected from critical errors and would raise an exception. This is done to keep simplicity and follow &quot;let it fall&quot; pattern of the language, thinking that these kind of errors should be logged and then manually investigated by a developer.But if that's not enough for you, it is possible to register handler via with_compensation_error_handler/2. When it's registered, compensations are wrapped in a try..catch block and then it's error handler responsibility to take care about further actions. Few solutions you might want to try:Send notification to a Slack channel about need of manual resolution;Retry compensation;Spawn a new supervised process that would retry compensation and return an error in the Sage. (Useful when you have connection issues that would be resolved at some point in future.)Logging for compensation errors is pretty verbose to drive the attention to the problem from system maintainers.","ref":"readme.html#critical-error-handling","title":"Sage - Critical Error Handling","type":"extras"},{"doc":"Sage does its best to make sure the final callback is executed even if there is a program bug in the code. This guarantee simplifies integration with job processing queues, you can read more about it at GenTask Readme.If an error is raised within the finally/2 hook, it gets logged and ignored. Follow the simple rule - everything that is on your critical path should be a Sage transaction.","ref":"readme.html#finally-2-hook","title":"Sage - finally/2 hook","type":"extras"},{"doc":"Sage allows you to set a tracer module which is called on each step of the execution flow (before and after transactions and/or compensations). It could be used to report metrics on the execution flow.If an error is raised within tracing function, it's getting logged and ignored.VisualizationsIn order to make it easier to understand what flow you should expect, here are a few additional examples:Retries[T1] -&gt; [T2] -&gt; [T3 has an error] ↓ [C2 retries] &lt;- [C3] ↓ [T2] -&gt; [T3]Circuit breaker[T1] -&gt; [T2 has an error] ↓ [C2 circuit breaker] -&gt; [T3]Async transactions[T1] -&gt; [T2 async] -↓ [T3 async] -&gt; [await for T2 and T3 before non-async operation] -&gt; [T4]Error in async transaction (notice: both async operations are awaited and then compensated)[T1] -&gt; [T2 async with error] -↓ [T3 async] -&gt; [await for T2 and T3 before non-async operation] ↓ [C1] &lt;- [C2] &lt;- [C3]","ref":"readme.html#tracing-and-measuring-sage-execution-steps","title":"Sage - Tracing and measuring Sage execution steps","type":"extras"},{"doc":"The package can be installed by adding sage to your list of dependencies in mix.exs:def deps do [ {:sage, &quot;~&gt; 0.4.0&quot;} ] endDocumentation can be found at https://hexdocs.pm/sage.LicenseSee LICENSE.md.CreditsSome implementation ideas were taken from Ecto.Multi module originally implemented by @michalmuskala and gisla by @mrallen1 which implements Sagas pattern for Erlang.Sagas idea have origins from whitepaper written in 80's. There are more recent work - Compensating Transactions which is part of Azure Architecture Guidelines.Thanks toJosh Forisha for letting me to use this awesome project name on hex.pm (he had a package with that name);@michalmuskala, @alco and @alecnmk for giving feedback and ideas along my way;all the Elixir community and Core Team. Guys, you are awesome ❤️.","ref":"readme.html#installation","title":"Sage - Installation","type":"extras"}]